@@ -1,4 +1,4 @@
-from sympy import preorder_traversal
+from sympy import preorder_traversal, latex
 
 
 class Query:
@@ -12,7 +12,7 @@ class Query:
         kwargs : keyword arguments
             Specifies what query to perform.
 
-            Allowed keys (only ONE at a time is allowed):
+            Keys (only ONE at a time is allowed):
 
                 'type'
                     Value can be any type.
@@ -48,50 +48,40 @@ class Query:
 
         self.tests = []
 
+        negate = kwargs.get('negate', False)
+
         if 'type' in kwargs:
-            self.tests.append(lambda e: type(e) == kwargs['type'])
+            self.tests.append(IsType(kwargs['type'], negate))
         elif 'isinstance' in kwargs:
-            # Pycharm reports an error here, but this is just a PyCharm bug.
-            # See: https://stackoverflow.com/questions/56493140/parameterized-generics-cannot-be-used-with-class-or-instance-checks
-            self.tests.append(lambda e: isinstance(e, kwargs['isinstance']))
+            self.tests.append(IsInstance(kwargs['isinstance'], negate))
         elif 'args' in kwargs:
-            args = kwargs['args']
-            if not isinstance(args, tuple) and not isinstance(args, list):
-                args = tuple(args)
-            def test_args(e):
-                if e.is_Atom: # atoms have no args
-                    return False
-                return all(arg in e.args for arg in args) and len(args) == len(e.args)
-            self.tests.append(test_args)
+            self.tests.append(ArgsEquals(kwargs['args'], negate))
         elif 'args__contains' in kwargs:
-            args = kwargs['args__contains']
-            if not isinstance(args, tuple) and not isinstance(args, list):
-                args = tuple(args)
-            def test_args__contains(e):
-                if e.is_Atom: # atoms have no args
-                    return False
-                return all(arg in e.args for arg in args)
-            self.tests.append(test_args__contains)
+            self.tests.append(ArgsContains(kwargs['args__contains'], negate))
         elif 'expr' in kwargs:
-            self.tests.append(
-                lambda e: e == kwargs['expr']
-            )
+            self.tests.append(ExprEquals(kwargs['expr'], negate))
+        elif 'latex' in kwargs:
+            self.tests.append(LatexEquals(kwargs['latex'], negate))
+        elif 'latex__contains' in kwargs:
+            self.tests.append(LatexContains(kwargs['latex__contains'], negate))
         elif 'test' in kwargs:
-            test = kwargs['test']
-            if not callable(test):
-                raise TypeError('{} is not callable.'.format(repr(test)))
-            self.tests.append(test)
+            self.tests.append(Predicate(kwargs['test'], negate))
         elif 'tests' in kwargs:
             tests = kwargs['tests']
-            err_msg = 'tests must be an iterable of callables.'
+            valid_tests = []
             try:
                 iter(tests)
                 for t in tests:
                     if not callable(t):
-                        raise ValueError(err_msg)
+                        raise ValueError
+                    if not isinstance(t, Predicate):
+                        t = Predicate(t)
+                    if negate:
+                        t = t.negated()
+                    valid_tests.append(t)
             except:
-                raise ValueError(err_msg)
-            self.tests = tests
+                raise ValueError('tests must be an iterable of callables.')
+            self.tests = valid_tests
         else:
             raise AssertionError('This should not happen.')
 
@@ -116,31 +106,130 @@ class Query:
         return Query(tests=tests)
 
     def __repr__(self):
-        return str(self.kwargs)
+        return repr(self.kwargs)
 
     def _validate_keywords(self, kwargs):
-        allowed_keywords = [
+        allowed_unique_keywords = [
             'type',  # tests if type are exactly equal
             'isinstance',  # tests if type is same or child class
             'expr',  # tests if the subexpr exactly matches
             'args',  # tests if args exactly match
             'args__contains',  # tests if args contain all of the given tuple items
+            'latex',  # tests is latex representation of subexpression matches
+            'latex__contains',  # tests if latex representation of subexpression contains value
             'test',  # user-defined matching test
             'tests',  # give initial set of test functions
         ]
+        allowed_other_keywords = [
+            'negate'
+        ]
         num_keywords_found = 0
         for key in kwargs:
-            if not key in allowed_keywords:
+            if key not in allowed_unique_keywords and key not in allowed_other_keywords:
                 raise ValueError('Invalid keyword in Query constructor: %s' % key)
-            num_keywords_found += 1
+            if key in allowed_unique_keywords:
+                num_keywords_found += 1
         if num_keywords_found == 0:
             raise ValueError('No query parameters defined.')
         if num_keywords_found > 1:
             raise ValueError('Only one query parameter may be set.')
 
 
+class Predicate:
+
+    def __init__(self, test, negate=False):
+        if not callable(test):
+            raise TypeError('Predicate argument must be callable!')
+        self._negate = negate
+        self._test = test
+
+    def __call__(self, expr, *args, **kwargs):
+        if self._negate:
+            return not self._test(expr, *args, **kwargs)
+        return self._test(expr, *args, **kwargs)
+
+    def negated(self):
+        return type(self)(
+            lambda e, *args, **kwargs: not self._test(e, *args, **kwargs),
+            negate=not self._negate
+        )
+
+    def __repr__(self):
+        if self._negate:
+            return 'not {}'.format(type(self).__name__)
+        return type(self).__name__
+
+
+class IsType(Predicate):
+
+    def __init__(self, the_type, negate=False):
+        super(IsType, self).__init__(lambda e: type(e) == the_type, negate)
+
+
+class IsInstance(Predicate):
+
+    def __init__(self, parent_type, negate=False):
+        super(IsInstance, self).__init__(lambda e: isinstance(e, parent_type), negate)
+
+
+class ExprEquals(Predicate):
+
+    def __init__(self, expr, negate=False):
+        super(ExprEquals, self).__init__(lambda e: e == expr, negate)
+
+
+class ArgsEquals(Predicate):
+
+    def __init__(self, args, negate=False):
+        if not isinstance(args, tuple) and not isinstance(args, list):
+            args = tuple(args)
+
+        def test_args(e):
+            if e.is_Atom:  # atoms have no args
+                return False
+            return all(arg in e.args for arg in args) and len(args) == len(e.args)
+
+        super(ArgsEquals, self).__init__(test_args, negate)
+
+
+class ArgsContains(Predicate):
+
+    def __init__(self, some_args, negate=False):
+        if not isinstance(some_args, tuple) and not isinstance(some_args, list):
+            some_args = tuple(some_args)
+
+        def test_args__contains(e):
+            if e.is_Atom:  # atoms have no args
+                return False
+            return all(arg in e.args for arg in some_args)
+
+        super(ArgsContains, self).__init__(test_args__contains, negate)
+
+
+class LatexEquals(Predicate):
+
+    def __init__(self, latex_str, negate=False):
+        def test_latex(e):
+            latex_repr = _searchable_latex_str(latex(e))
+            return latex_repr == _searchable_latex_str(latex_str)
+
+        super(LatexEquals, self).__init__(test_latex, negate)
+
+
+class LatexContains(Predicate):
+
+    def __init__(self, latex_str, negate=False):
+        def test_latex_contains(e):
+            latex_repr = _searchable_latex_str(latex(e))
+            return _searchable_latex_str(latex_str) in latex_repr
+
+        super(LatexContains, self).__init__(test_latex_contains, negate)
+
+
 class QueryResult:
 
+    # TODO: intersection and union of query results
+
     def __init__(self, expr_list=None):
         self._expr_list = expr_list or []
         self._counter = None
@@ -163,6 +252,9 @@ class QueryResult:
     def extend(self, expr_list):
         self._expr_list.extend(expr_list)
 
+    def as_set(self):
+        return set(self._expr_list)
+
     def __iter__(self):
         self._counter = 0
         return self
@@ -311,3 +403,12 @@ def get_epath(subexpr, containing_expr):
     if len(paths) == 0:
         raise NotFoundException
     return paths[0]
+
+
+def get_level(expr, level):
+    # TODO, as helper for sympy.use
+    raise NotImplementedError
+
+
+def _searchable_latex_str(latex_repr):
+    return latex_repr.replace(' ', '').replace(r'\left', '').replace(r'\right', '')
